require 'gosu'
require 'opengl'
require 'glu'
include Gl, Glu

class Gosu::GLTexInfo
	def tex_coord_2d(x, y)
		tex_x = self.left + x * (self.right - self.left)
		tex_y = (self.bottom + y * (self.top - self.bottom))
		glTexCoord2d(tex_x, tex_y)
	end
	
	def bottom_left;	tex_coord_2d(0.0, 0.0);	end
	def bottom_right;	tex_coord_2d(1.0, 0.0);	end
	def top_left;	tex_coord_2d(0.0, 1.0);	end
	def top_right;	tex_coord_2d(1.0, 1.0);	end
end

class Gosu::Image
	# x, y and z are relative to the center of the image
	def draw_3d(x, y, z, options = {})
		width, height = self.width, self.height
		x, y, w, h = 0.0, 0.0, 1.0, 1.0

		if options.has_key?(:rect)
			# rect uses the convention [x, y, w, h], expressed in opengl (from 0.0 to 1.0, origin on bottom-left corner)
			x, y, w, h = options[:rect][0].to_f, options[:rect][1].to_f, options[:rect][2].to_f, options[:rect][3].to_f
			x = 0.0 if x < 0.0 or x > 1.0
			y = 0.0 if y < 0.0 or y > 1.0
			w = 0.0 if w < 0.0 or w > 1.0
			h = 0.0 if h < 0.0 or h > 1.0
			width, height = w * self.width, h * self.height
		end
		
		glBindTexture(GL_TEXTURE_2D, self.gl_tex_info.tex_name)
		glEnable(GL_ALPHA_TEST)
		glPushMatrix
		glTranslate(x, y, z)
		glScale(width, height, 1.0)		
		
		if options.has_key?(:rot)
			# rot uses the convention [x, y, z], expressed in degrees
			glRotate(options[:rot][0], 1, 0, 0)
			glRotate(options[:rot][1], 0, 1, 0)
			glRotate(options[:rot][2], 0, 0, 1)
		end

		glBegin(GL_QUADS)
			self.gl_tex_info.tex_coord_2d(x, y); glVertex3f(-0.5, -0.5, 0.0)
			self.gl_tex_info.tex_coord_2d(x + w, y); glVertex3f(0.5, -0.5, 0.0)
			self.gl_tex_info.tex_coord_2d(x + w, y + h); glVertex3f(0.5, 0.5, 0.0)
			self.gl_tex_info.tex_coord_2d(x, y + h); glVertex3f(-0.5, 0.5, 0.0)
		glEnd
		glPopMatrix
		glDisable(GL_ALPHA_TEST)
	end
end

module RetroRender
	class Vector3
		attr_accessor :x, :y, :z
		# @param [Float] x
		# @param [Float] y
		# @param [Float] z	
		def initialize(x = 0.0, y = 0.0, z = 0.0)
			@x, @y, @z = x, y, z
		end
	end
	
	class OpenGLContext
		def initialize
			glEnable(GL_TEXTURE_2D)
			glEnable(GL_DEPTH_TEST)
			glClearColor(0.0, 1.0, 0.0, 0.0)
		end
		
		def clear
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
		end
	end
	
	class Camera
		attr_accessor :position, :target, :fovy, :ratio, :near, :far, :angles, :distance
		# == ARGS : 
		# Hash that can take the following keys :
		# @param [Symbol] :mode (defaults to: :fps)
		# @param [Vector3] :position (defaults to: Vector3.new(0.0, 0.0, 0.0))
		# @param [Vector3] :target (defaults to: Vector3.new(0.0, 0.0, 0.0))
		# @param [Vector3] :angles (defaults to: Vector3.new(0.0, 0.0, 0.0))
		# @param [Float] :fovy (defaults to: 45.0)
		# @param [Float] :ratio (defaults to: 1.333)
		# @param [Float] :near (defaults to: 0.1)
		# @param [Float] :far (defaults to: 1000.0)
		# @param [Float] :distance (defaults to: 64.0 - distance between the position and the target)
		def initialize(args = {})
			@mode = args.has_key?(:mode) ? args[:mode] : :fps
			@position = args.has_key?(:position) ? args[:position] : Vector3.new
			@target = args.has_key?(:target) ? args[:target] : Vector3.new
			@angles = args.has_key?(:angles) ? args[:angles] : Vector3.new
			@fovy = args.has_key?(:fovy) ? args[:fovy] : 45.0
			@ratio = args.has_key?(:ratio) ? args[:ratio] : 1.333
			@near = args.has_key?(:near) ? args[:near] : 0.1
			@far = args.has_key?(:far) ? args[:far] : 1000.0
			@distance = args.has_key?(:distance) ? args[:distance] : 64.0
			@deg_to_rad = Math::PI / 180.0
		end
		
		def move_forward(velocity)
			if @mode == :fps
				@position.x += velocity * Math::cos(@angles.x * @deg_to_rad)
				@position.z += velocity * Math::sin(@angles.x * @deg_to_rad)
			else
				@target.x += velocity * Math::cos(@angles.x * @deg_to_rad)
				@target.z += velocity * Math::sin(@angles.x * @deg_to_rad)
			end
		end
		
		def move_left(velocity)
			if @mode == :fps
				@position.x += velocity * Math::cos((@angles.x - 90.0) * @deg_to_rad)
				@position.z += velocity * Math::sin((@angles.x - 90.0) * @deg_to_rad)
			else
				@target.x += velocity * Math::cos((@angles.x - 90.0) * @deg_to_rad)
				@target.z += velocity * Math::sin((@angles.x - 90.0) * @deg_to_rad)
			end
		end

		def move_right(velocity)
			if @mode == :fps
				@position.x -= velocity * Math::cos((@angles.x - 90.0) * @deg_to_rad)
				@position.z -= velocity * Math::sin((@angles.x - 90.0) * @deg_to_rad)
			else
				@target.x -= velocity * Math::cos((@angles.x - 90.0) * @deg_to_rad)
				@target.z -= velocity * Math::sin((@angles.x - 90.0) * @deg_to_rad)
			end
		end

		def move_backward(velocity)
			if @mode == :fps
				@position.x -= velocity * Math::cos(@angles.x * @deg_to_rad)
				@position.z -= velocity * Math::sin(@angles.x * @deg_to_rad)
			else
				@target.x -= velocity * Math::cos(@angles.x * @deg_to_rad)
				@target.z -= velocity * Math::sin(@angles.x * @deg_to_rad)
			end
		end
		
		def rotate(axis, value)
			if axis == :x
				@angles.x += value
			elsif axis == :y
				@angles.y += value
			elsif axis == :z
				@angles.z += value
			else
				raise("axis must be a symbol as :x, :y or :z")
			end
		end

		def update
			if @mode == :fps
				@target.x = @position.x + @distance * Math::cos(@angles.y * @deg_to_rad)
				@target.z = @position.z + @distance * Math::cos(@angles.y * @deg_to_rad)				
				@target.x = @position.x + @distance * Math::cos(@angles.x * @deg_to_rad)
				@target.y = @position.y + @distance * Math::sin(@angles.y * @deg_to_rad)
				@target.z = @position.z + @distance * Math::sin(@angles.x * @deg_to_rad)
			else
				@position.x = @target.x - @distance * Math::cos(@angles.y * @deg_to_rad)
				@position.z = @target.z - @distance * Math::cos(@angles.y * @deg_to_rad)
				@position.x = @target.x - @distance * Math::cos(@angles.x * @deg_to_rad)
				@position.y = @target.y - @distance * Math::sin(@angles.y * @deg_to_rad)
				@position.z = @target.z - @distance * Math::sin(@angles.x * @deg_to_rad)
			end
		end
		
		# Sets the OpenGL matrices according to the camera
		def look
			glMatrixMode(GL_PROJECTION)
			glLoadIdentity
			gluPerspective(@fovy, @ratio, @near, @far)
			glMatrixMode(GL_MODELVIEW)
			glLoadIdentity
			gluLookAt(@position.x, @position.y, @position.z, @target.x, @target.y, @target.z, 0, 1, 0)
		end
	end
end